@@ -3,9 +3,10 @@ from jax.config import config as jaxconfig
 jaxconfig.update("jax_enable_x64", True)
 jaxconfig.update("jax_platform_name", "cpu")
 
-from .base_cov import Covariance, Add, Mul, Pow
-from .util import stabilize, mle, distance, DEFAULT_JITTER
-from .decomposition import DEFAULT_RANK, DEFAULT_METHOD
+__version__ = "1.1.0"
+
+from .base_cov import Covariance
+from .util import stabilize, mle, distance, Log
 from .cov import Matern32, Matern52, ExpQuad, Exponential, RatQuad
 from .inference import (
     compute_transform,
@@ -14,8 +15,6 @@ from .inference import (
     minimize_lbfgsb,
     compute_log_density_x,
     compute_conditional_mean,
-    DEFAULT_N_ITER,
-    DEFAULT_INIT_LEARN_RATE,
 )
 from .parameters import (
     compute_landmarks,
@@ -27,7 +26,6 @@ from .parameters import (
     compute_cov_func,
     compute_L,
     compute_initial_value,
-    DEFAULT_N_LANDMARKS,
 )
 from .derivatives import (
     gradient,
@@ -37,5 +35,4 @@ from .derivatives import (
 from .model import (
     DensityEstimator,
     FunctionEstimator,
-    DEFAULT_COV_FUNC,
 )
