@@ -1,4 +1,3 @@
-from warnings import warn
 from .cov import Matern52
 from .decomposition import DEFAULT_RANK, DEFAULT_METHOD
 from .inference import (
@@ -8,6 +7,7 @@ from .inference import (
     minimize_lbfgsb,
     compute_log_density_x,
     compute_conditional_mean,
+    compute_conditional_mean_y,
     DEFAULT_N_ITER,
     DEFAULT_INIT_LEARN_RATE,
     DEFAULT_JIT,
@@ -29,15 +29,20 @@ from .derivatives import (
     hessian,
     hessian_log_determinant,
 )
-from .util import DEFAULT_JITTER
+from .util import (
+    DEFAULT_JITTER,
+    vector_map,
+    Log,
+)
 
 
 DEFAULT_COV_FUNC = Matern52
 
+logger = Log()
 
 class BaseEstimator:
     R"""
-    Base class for mellon estimators.
+    Base class for the mellon estimators.
     """
 
     def __init__(
@@ -45,7 +50,6 @@ class BaseEstimator:
         cov_func_curry=DEFAULT_COV_FUNC,
         n_landmarks=DEFAULT_N_LANDMARKS,
         rank=DEFAULT_RANK,
-        method=DEFAULT_METHOD,
         jitter=DEFAULT_JITTER,
         optimizer=DEFAULT_OPTIMIZER,
         n_iter=DEFAULT_N_ITER,
@@ -63,7 +67,6 @@ class BaseEstimator:
         self.cov_func_curry = cov_func_curry
         self.n_landmarks = n_landmarks
         self.rank = rank
-        self.method = method
         self.jitter = jitter
         self.landmarks = landmarks
         self.nn_distances = nn_distances
@@ -84,9 +87,7 @@ class BaseEstimator:
             f"cov_func_curry={self.cov_func_curry}, "
             f"n_landmarks={self.n_landmarks}, "
             f"rank={self.rank}, "
-            f"method='{self.method}', "
             f"jitter={self.jitter}, "
-            f"n_iter={self.n_iter}, "
             f"landmarks={self.landmarks}, "
         )
         if self.nn_distances is None:
@@ -111,11 +112,13 @@ class BaseEstimator:
     def _compute_landmarks(self):
         x = self.x
         n_landmarks = self.n_landmarks
+        logger.info(f"Computing {n_landmarks:,} landmarks with k-means clustering.")
         landmarks = compute_landmarks(x, n_landmarks=n_landmarks)
         return landmarks
 
     def _compute_nn_distances(self):
         x = self.x
+        logger.info('Computing nearest neighbor distances.')
         nn_distances = compute_nn_distances(x)
         return nn_distances
 
@@ -129,25 +132,42 @@ class BaseEstimator:
         cov_func_curry = self.cov_func_curry
         ls = self.ls
         cov_func = compute_cov_func(cov_func_curry, ls)
+        logger.info("Using covariance function %s.", str(cov_func))
         return cov_func
 
     def _compute_L(self):
         x = self.x
         cov_func = self.cov_func
         landmarks = self.landmarks
+        n_landmarks = landmarks.shape[0]
         rank = self.rank
         method = self.method
         jitter = self.jitter
+        if isinstance(rank, float) and method != 'fixed':
+            logger.info(
+                f'Computing rank reduction using "{method}" method '
+                f"retaining > {rank:.2%} of variance."
+            )
+        else:
+            logger.info(
+                f'Computing rank reduction to rank {rank} using "{method}" method.'
+            )
         L = compute_L(
             x, cov_func, landmarks=landmarks, rank=rank, method=method, jitter=jitter
         )
         new_rank = L.shape[1]
-        if new_rank > (.8*rank):
-            warn(
-                f'Shallow rank reduction from {rank:,} to {new_rank:,} '
-                'indicates underrepresentation by landmarks. Consider '
-                'increasing n_landmarks!'
+        if not (
+            type(rank) is int
+            and rank == n_landmarks
+            or type(rank) is float
+            and rank == 1.0
+        ) and method != 'fixed' and new_rank > (rank * 0.8 * n_landmarks):
+            logger.warning(
+                f"Shallow rank reduction from {n_landmarks:,} to {new_rank:,} "
+                "indicates underrepresentation by landmarks. Consider "
+                "increasing n_landmarks!"
             )
+        logger.info(f"Using rank {new_rank:,} covariance representation.")
         return L
 
     def _run_inference(self):
@@ -156,6 +176,7 @@ class BaseEstimator:
         n_iter = self.n_iter
         init_learn_rate = self.init_learn_rate
         optimizer = self.optimizer
+        logger.info("Running inference using %s.", optimizer)
         if optimizer == "adam":
             results = minimize_adam(
                 function,
@@ -179,12 +200,14 @@ class BaseEstimator:
                 results.loss,
             ]
         else:
-            raise ValueError(
+            error = ValueError(
                 f"Unknown optimizer {optimizer}. You can use .loss_func and "
                 ".initial_value as loss function and initial state for an "
                 "external optimization. Write optimal state to "
                 ".pre_transformation to enable prediction with .predict()."
             )
+            logger.error(error)
+            raise error
 
     def _prepare_attribute(self, attribute):
         R"""
@@ -422,19 +445,19 @@ class DensityEstimator(BaseEstimator):
         jit=DEFAULT_JIT,
     ):
         super().__init__(
-            cov_func_curry = cov_func_curry,
-            n_landmarks = n_landmarks,
-            rank = rank,
-            method = method,
-            jitter = jitter,
-            landmarks = landmarks,
-            nn_distances = nn_distances,
-            mu = mu,
-            ls = ls,
-            ls_factor = ls_factor,
-            cov_func = cov_func,
-            L = L,
+            cov_func_curry=cov_func_curry,
+            n_landmarks=n_landmarks,
+            rank=rank,
+            jitter=jitter,
+            landmarks=landmarks,
+            nn_distances=nn_distances,
+            mu=mu,
+            ls=ls,
+            ls_factor=ls_factor,
+            cov_func=cov_func,
+            L=L,
         )
+        self.method = method
         self.optimizer = optimizer
         self.n_iter = n_iter
         self.init_learn_rate = init_learn_rate
@@ -481,13 +504,9 @@ class DensityEstimator(BaseEstimator):
             string += "initial_value=None, "
         else:
             string += "initial_value=initial_value, "
-        string += (
-            f"jit={self.jit}"
-            ")"
-        )
+        string += f"jit={self.jit}" ")"
         return string
 
-
     def _compute_d(self):
         x = self.x
         d = compute_d(x)
@@ -528,42 +547,6 @@ class DensityEstimator(BaseEstimator):
         loss_func = compute_loss_func(nn_distances, d, transform, k)
         return loss_func
 
-    def _run_inference(self):
-        function = self.loss_func
-        initial_value = self.initial_value
-        n_iter = self.n_iter
-        init_learn_rate = self.init_learn_rate
-        optimizer = self.optimizer
-        if optimizer == "adam":
-            results = minimize_adam(
-                function,
-                initial_value,
-                n_iter=n_iter,
-                init_learn_rate=init_learn_rate,
-                jit=self.jit,
-            )
-            self.pre_transformation = results.pre_transformation
-            self.opt_state = results.opt_state
-            self.losses = results.losses
-        elif optimizer == "L-BFGS-B":
-            results = minimize_lbfgsb(
-                function,
-                initial_value,
-                jit=self.jit,
-            )
-            self.pre_transformation = results.pre_transformation
-            self.opt_state = results.opt_state
-            self.losses = [
-                results.loss,
-            ]
-        else:
-            raise ValueError(
-                f"Unknown optimizer {optimizer}. You can use .loss_func and "
-                ".initial_value as loss function and initial state for an "
-                "external optimization. Write optimal state to "
-                ".pre_transformation to enable prediction with .predict()."
-            )
-
     def _set_log_density_x(self):
         pre_transformation = self.pre_transformation
         transform = self.transform
@@ -577,8 +560,14 @@ class DensityEstimator(BaseEstimator):
         mu = self.mu
         cov_func = self.cov_func
         jitter = self.jitter
+        logger.info("Computing predictive function.")
         log_density_func = compute_conditional_mean(
-            x, landmarks, log_density_x, mu, cov_func, jitter=jitter,
+            x,
+            landmarks,
+            log_density_x,
+            mu,
+            cov_func,
+            jitter=jitter,
         )
         self.log_density_func = log_density_func
 
@@ -717,10 +706,14 @@ class DensityEstimator(BaseEstimator):
         """
         if self.x is not None and self.x is not x:
             message = "self.x has been set already, but is not equal to the argument x."
-            raise ValueError(message)
+            error = ValueError(message)
+            logger.error(error)
+            raise error
         if self.x is None and x is None:
             message = "Required argument x is missing and self.x has not been set."
-            raise ValueError(message)
+            error = ValueError(message)
+            logger.error(error)
+            raise error
         if x is None:
             x = self.x
 
@@ -782,9 +775,6 @@ class FunctionEstimator(BaseEstimator):
         k(x, y) :math:`\rightarrow` float. If None, automatically generates the covariance
         function cov_func = cov_func_curry(ls). Defaults to None.
     :type cov_func: function or None
-    :param L: A matrix such that :math:`L L^\top \approx K`, where :math:`K` is the covariance matrix.
-        If None, automatically computes L. Defaults to None.
-    :type L: array-like or None
     :param sigma: The white moise standard deviation. Defaults to 0.
     :type sigma: float
     :ivar n_landmarks: The number of landmark points.
@@ -825,27 +815,22 @@ class FunctionEstimator(BaseEstimator):
         ls=None,
         ls_factor=1,
         cov_func=None,
-        L=None,
         sigma=0,
     ):
         super().__init__(
-            cov_func_curry = cov_func_curry,
-            n_landmarks = n_landmarks,
-            rank = rank,
-            method = method,
-            jitter = jitter,
-            landmarks = landmarks,
-            nn_distances = nn_distances,
-            mu = mu,
-            ls = ls,
-            ls_factor = ls_factor,
-            cov_func = cov_func,
-            L = L,
+            cov_func_curry=cov_func_curry,
+            n_landmarks=n_landmarks,
+            rank=rank,
+            jitter=jitter,
+            landmarks=landmarks,
+            nn_distances=nn_distances,
+            mu=mu,
+            ls=ls,
+            ls_factor=ls_factor,
+            cov_func=cov_func,
         )
-        self.y = None
         self.sigma = sigma
 
-
     def _prepare_attribute(self, attribute):
         R"""
         If self.attribute is None, sets self.attribute to the value of its
@@ -861,10 +846,7 @@ class FunctionEstimator(BaseEstimator):
         value = function()
         setattr(self, attribute, value)
 
-    def _set_y(self, y):
-        self.y = y
-
-    def prepare_inference(self, x, y):
+    def prepare_inference(self, x):
         R"""
         Set all attributes in preparation. It is not necessary to call this
         function before calling fit.
@@ -878,15 +860,12 @@ class FunctionEstimator(BaseEstimator):
         :rtype: function, array-like
         """
         self._set_x(x)
-        self._set_y(y)
         self._prepare_attribute("nn_distances")
         self._prepare_attribute("ls")
         self._prepare_attribute("cov_func")
         self._prepare_attribute("landmarks")
-        self._prepare_attribute("L")
         return
 
-
     def compute_conditional(self, x=None, y=None):
         R"""
         Compute and return the conditional mean function.
@@ -906,21 +885,22 @@ class FunctionEstimator(BaseEstimator):
             raise ValueError(message)
         if x is None:
             x = self.x
-        if self.y is not None and self.y is not y:
-            message = "self.y has been set already, but is not equal to the argument y."
-            raise ValueError(message)
-        if self.y is None and y is None:
-            message = "Required argument y is missing and self.y has not been set."
-            raise ValueError(message)
         if y is None:
-            y = self.y
+            message = "Required argument y is missing."
+            raise ValueError(message)
         landmarks = self.landmarks
         mu = self.mu
         cov_func = self.cov_func
         sigma = self.sigma
         jitter = self.jitter
         conditional = compute_conditional_mean(
-            x, landmarks, y, mu, cov_func, sigma, jitter=jitter,
+            x,
+            landmarks,
+            y,
+            mu,
+            cov_func,
+            sigma,
+            jitter=jitter,
         )
         self.conditional = conditional
         return conditional
@@ -940,7 +920,7 @@ class FunctionEstimator(BaseEstimator):
         :rtype: Object
         """
 
-        self.prepare_inference(x, y)
+        self.prepare_inference(x)
         self.compute_conditional(x, y)
         return self
 
@@ -960,7 +940,7 @@ class FunctionEstimator(BaseEstimator):
         Compute the conditional mean and return the smoothed function values
         at the points x.
 
-        :param x: The training instances to estimate density function.
+        :param x: The training instances to estimate function.
         :type x: array-like
         :param y: The training function values on cell states.
         :type y: array-like
@@ -970,3 +950,41 @@ class FunctionEstimator(BaseEstimator):
 
         self.fit(x, y)
         return self.predict(x)
+
+    def multi_fit_predict(self, x=None, Y=None, Xnew=None):
+        R"""
+        Compute the conditional mean and return the smoothed function values
+        at the points Xnew for each line of values in Y.
+
+        :param x: The training instances to estimate density function.
+        :type x: array-like
+        :param y: The training function values on cell states.
+        :type y: array-like
+        :param Xnew: The new data to predict.
+        :type Xnew: array-like
+        :return: condition_mean - The conditional mean function value at each test point in x.
+        :rtype: array-like
+        """
+
+        if Xnew is None:
+            Xnew = x
+
+        self.prepare_inference(x)
+
+        landmarks = self.landmarks
+        mu = self.mu
+        cov_func = self.cov_func
+        sigma = self.sigma
+        jitter = self.jitter
+
+        conditional = compute_conditional_mean_y(
+            x,
+            landmarks,
+            Xnew,
+            mu,
+            cov_func,
+            sigma,
+            jitter=jitter,
+        )
+
+        return vector_map(conditional, Y)
